-[ c library "bgrects.c", header "bgrects.h" ]

-[ headers, begin ]
	!stdlib !stdio !string
	parsedoc style repstr
-[ end ]

-[private] typedef struct { int x1,y1; int x2,y2; } rect_t;

-[private] typedef struct
	{ int color; rect_t area; } fill_t; 

-[ fbuffer_t : private buffer [500] of fill_t ]

void gen_bgrects(grid_t *G, strpairlist_t *vars)
{
	fbuffer_t *buffer;
	int nrects;
	fill_t *rects;
	int i,j;

	buffer= fbuffer_new();
	-[ generate cell background rectangles ]
	rects= fbuffer_collectda(buffer, &nrects);

	-[ quicksort array rects = fill_t[nrects] , begin ]
		-[ compare X1 to X2 ]
			result= X1->color - X2->color;
	-[end] 

	buffer= fbuffer_new();

	for(i=0;i<nrects;i++) {
		rect_t *R;
		int nr;
		int C;
		for(j=i+1;j<nrects;j++) 
			if (rects[j].color!=rects[i].color)
				break;
		-[ R= allocate array j-i ]
		nr= 0;
		C= rects[i].color;
		while(i<j) 
			R[nr++]= rects[i++].area;
		i= j-1;

		R= merge_rectangles(R, &nr);
		-[ put R into buffer ]
		free(R);
	}
}


=[ generate cell background rectangles ]
	for(i=0;i<G->nrows;i++)
		for(j=0;j<G->ncols;j++) {
			xcell_t *X;
			fill_t T;
			X= &G->cells[i][j];
			if (X->kind==ck_merged) continue;
			T.color= X->bgcolor;
			T.area.y1= i; T.area.x1= j;
			T.area.y2= i+X->rowspan;
			T.area.x2= j+X->colspan;
			fbuffer_append(buffer,T);
		}

=[ put R into buffer ] {
	int k;
	for(k=0;k<nr;k++) {
		fill_t F;
		F.area= R;
		F.color= C;
		fbuffer_append(buffer,F);
	}
}

=[]

-[ import bitvector routines in private scope ]
-[ import bitmatrix routines in private scope ]
-[ import allpaths function in private scope ]

-[private] rect_t *merge_rectangles(rect_t *input,int *n)
{
	rbuffer_t *obuffer;
	int N;
	int i,j;
	bitmatrix_t *neighbours;
	bitvector_t *visited;
	N= *n;

	obuffer= rbuffer_new();

	neighbours= bitmatrix_new(N,N);

	for(i=0;i<N;i++) {
		bitmatrix_set(neighbours,i,i,1);
		for(j=i+1;j<N;j++)
			if (are_neighbours,input+i,input+j)) {
				bitmatrix_set(neighbours,i,j,1);
				bitmatrix_set(neighbours,j,i,1);
			}
	}

	allpaths(neighbours);
	visited= bitvector_new(N);

	for(i=0;i<N;i++) {
		rbuffer_t *buffer;
		int nc;
		int j,k;
		rect_t *conn;

		if (bitvector_get(visited,i))
			continue; 
		buffer= rbuffer_new();

		rbuffer_append(buffer, input[i]);
		bitvector_set(visited,i);

		for(j=i+1;j<N;j++)  {
			if (bitvector_get(visited,j))
				continue;
			if (bitmatrix_get(neighbours, i,j)) {
				bitvector_set(visited,j);
				rbuffer_append(buffer, input[j]);
			}
		}
		conn= rbuffer_collectda(buffer,&nc);
		conn= merge_connected_rectangles(conn,&nc);
		for(k=0;k<nc;k++)
			rbuffer_append(obuffer, conn[k]);
		free(conn);
	}
	free(visited);
	bitmatrix_free(neighbours);
	free(input);
	return obuffer_collectda(obuffer,n);
}

-[private] int are_neighbours(rect_t *A,rect_t *B)
{
	if (B->x1==A->x1 && B->x2==A->x2) {
		if (B->y2==A->y1 || B->y1=A->y2)
			return 1;
	}
	if (B->y1==A->y1 && B->y2==A->y2) {
		if (B->x2==A->x1 || B->x1=A->x2)
			return 1;
	}
	return 0;
}

-[private] rect_t *merge_connected_rectangles
	(rect_t *input, int *n)
{
	int N= *n;

}


