-[ c library "code.c", header "code.h" ]

static char header_code[]= {
-[ quote code, begin ]

namespace FX {
class FXAPI $CLASS : public FXScrollArea {
	FXDECLARE($CLASS)
protected:
	$CLASS();
private:
	$CLASS(const $CLASS&);
	$CLASS &operator=(const $CLASS&);

	FXApp* myapp;

private:
	int input_r;
	int input_c;

	FXFont *data_font;
	FXFont *header_font;
	int data_font_ascent;
	int header_font_ascent;
	int data_font_height;
	int header_font_height;
	int data_font_swidth;
	int header_font_swidth;

	FXTextField *editor;
private:
	typedef struct { int y; int x1; int x2; } xline_t;
	typedef struct { int x; int y1; int y2; } yline_t;
	typedef struct { int x1,y1; int x2,y2; } crect_t;
	typedef struct { crect_t area; unsigned color; } bgfill_def_t;
	typedef struct { crect_t area; FXColor color; } bgfill_act_t;

	typedef enum { $FIELD_ENUMS  } cell_storage_kind_t;
	typedef enum { cell_int, cell_str } cell_base_type_t;

	typedef enum
		{ ck_merged, ck_empty, ck_data, ck_header }
			cell_kind_t;

	typedef struct
		{ cell_kind_t kind; int halign, valign;
		   int rowspan, colspan;
		   int merge_r, merge_c; int editable;
			cell_base_type_t  base_type;
		   cell_storage_kind_t skind;
		   int idx1, idx2;
		   const char *format;
		   int show_if_zero;
		 } cell_def_t;

	typedef struct
		{ int cwidth, cheight;
		char *data; } cell_t;

	static const xline_t horz_line_defs[$NHORZLINES];
	static const yline_t vert_line_defs[$NVERTLINES];
	static const bgfill_def_t bgfill_defs[$NBGFILLS];
	static const cell_def_t cell_defs[$NROWS][$NCOLS];

	xline_t horz_lines[$NHORZLINES];
	yline_t vert_lines[$NVERTLINES];
	bgfill_act_t bgfills[$NBGFILLS];

	cell_t cells[$NROWS][$NCOLS];

	int row_heights[$NROWS];
	int col_widths[$NCOLS];
	int x_offsets[$NCOLS+1];
	int y_offsets[$NROWS+1];

	int x_pad, y_pad;

	void get_text_size(int bold,char *data, int *w,int *h);
	void get_single_line_text_size
		(FXFont* font,char*data,int *tw,int *th);

	int render_text_line
		(FXDC* dc,FXFont *font,int X, int Y,
		int halign, char *data, int avail_w);
	void render_text
		(FXDC* dc,crect_t area, int valign,int halign, int bold,
		char *data,int content_width,int content_height);

	int calculate_col_widths();
	int calculate_row_heights();
	void calculate_cell_geometry(int r,int c);
	void calculate_cell_content_geometries();
	void convert_to_actual_coordinates();

	int ceil(int a,int b);
	int classify_point(crect_t R, int x,int y);
	int rectangles_intersect(crect_t R1,crect_t R2);
	int xline_visible(crect_t R, xline_t L);
	int yline_visible(crect_t R,yline_t L);
	FXColor rgb24tofxcolor(unsigned int x);
	int fit_to_range(int x,int min,int max);
	int unformat_int(char *txt,const char *fmt);
	char *unformat_str(char *txt,const char *fmt);
	char* format_int(int v,const char *fmt);
	char *format_str(char *txt,const char *fmt);

	void make_cell_visible(int r,int c);
public:
	class boxed_int {
	public:
		int value;
		$CLASS* table;
		void set(int v);
		int row,col;
		void init($CLASS *tb,int r,int c);
		void start_edit();
		void serialize(char **buffer,size_t *len);
		void deserialize(char **buffer,size_t *len);
	};

	class boxed_str {
	public:
		char *value;
		$CLASS* table;
		void set(char *v);
		int row,col;
		void init($CLASS *tb,int r,int c);
		void start_edit();
		void serialize(char **buffer,size_t *len);
		void deserialize(char **buffer,size_t *len);
	};

	struct {
		$STORAGE_DECLARATIONS
	} S;

	typedef struct {
		$RAW_STORAGE_DECLARATIONS
	} raw_storage_t;

	$CELL_CHANGE_CALLBACK_DECLARATIONS

	void serialize_data(char **buf,size_t *len);
	void deserialize_data(char *buf, size_t len);

	static raw_storage_t* parse_data(char *buf,size_t len);
	static void unparse_data(raw_storage_t *V,
		char **buf,size_t*len);
	static void destroy_data(raw_storage_t *obj);

	int show_if_zero(int row,int col);

public:
	$CLASS(FXComposite* p,
		FXuint opts=FRAME_NORMAL,
		FXint x=0,FXint y=0,FXint w=0,FXint h=0);
	virtual ~$CLASS();

	enum { ID_UPDATECELL= FXScrollBar::ID_LAST+1, };

	enum { valign_top, valign_center, valign_bottom };
	enum { halign_left, halign_center, halign_right };

	virtual bool canFocus() const { return true; }
	virtual void create();
	virtual void detach();

	virtual int getContentWidth();
	virtual int getContentHeight();
	virtual FXint getDefaultWidth();
	virtual FXint getDefaultHeight();

	void begin_update();
	void end_update();

	virtual void layout();

	long onPaint(FXObject*,FXSelector,void*);
	long onClick(FXObject*,FXSelector,void*);
	long onCellChange(FXObject*,FXSelector,void*);

public:
	int col(const char* X);
	int editable;
	void start_edit(int r,int c);
	void set_cell_text(int r,int c,char *text);
	void set_cell_text_noupdate(int r,int c,char *text);
};

}
-[end]
0};

static char source_code[]= {
-[ quote code, begin ]
#include <fx.h>
#include "$FILENAME.h"

FXDEFMAP($CLASS) $CLASSMessageMap[]={
	FXMAPFUNC(SEL_PAINT,0,$CLASS::onPaint),
	FXMAPFUNC(SEL_LEFTBUTTONPRESS, 0, $CLASS::onClick),
	FXMAPFUNC(SEL_COMMAND, $CLASS::ID_UPDATECELL,
			$CLASS::onCellChange),
};

FXIMPLEMENT($CLASS,FXScrollArea,
	$CLASSMessageMap,ARRAYNUMBER($CLASSMessageMap))




void $CLASS::create() { 
	FXScrollArea::create();

	data_font= new FXFont(myapp,"Courier", 14);
	data_font->create();
	data_font_ascent= data_font->getFontAscent();
	data_font_height= data_font->getFontHeight();
	data_font_swidth= data_font->getCharWidth(' ');

	header_font= new FXFont(myapp, "Courier", 14,FXFont::Bold);
	header_font->create();
	header_font_ascent= header_font->getFontAscent();
	header_font_height= header_font->getFontHeight();
	header_font_swidth= header_font->getCharWidth(' ');

	calculate_cell_content_geometries();
	calculate_col_widths();
	calculate_row_heights();
	convert_to_actual_coordinates();

	vertical->setLine(data_font_height);
	horizontal->setLine(data_font_swidth);
}

void $CLASS::serialize_data(char **buf,size_t *len) {
	size_t *L;
	char **B;

	char *buffer;
	size_t buffer_len;

	buffer= 0;
	buffer_len= 0;
	L= &buffer_len;
	B= &buffer;

again:
	$SERIALIZATION
	if (buffer) return ;
	buffer= (char*) malloc(buffer_len);
	*buf= buffer;
	*len= buffer_len;
	buffer_len= 0;
	goto again;
}

static void unparse_int(char **buffer, size_t *len, int value)
{
	if (*buffer) {
		memcpy(*buffer, (void*) &value, sizeof(value));
		*buffer += sizeof(value);
	}
	*len += sizeof(value);
}
static void unparse_str(char **buffer, size_t* len,char *value) {
	int L;
	if (value && value[0]) {
		L= strlen(value);
	} else L= 0;
	unparse_int(buffer, len, L);
	if (L) {
		if (*buffer) {
			memcpy(*buffer, value, L);
			*buffer+= L;
		}
		*len+= L;
	}
}


void $CLASS::unparse_data(raw_storage_t *V,char **buf,size_t *len) {
	size_t *L;
	char **B;

	char *buffer;
	size_t buffer_len;

	buffer= 0;
	buffer_len= 0;
	L= &buffer_len;
	B= &buffer;

again:
	$RAW_UNPARSING_CODE
	if (buffer) return ;
	buffer= (char*) malloc(buffer_len);
	*buf= buffer;
	*len= buffer_len;
	buffer_len= 0;
	goto again;
}

void $CLASS::boxed_int::serialize(char **buffer,size_t *len) {
	unparse_int(buffer,len,value);
}

void $CLASS::boxed_str::serialize(char **buffer, size_t *len)
{
	unparse_str(buffer,len,value);
}


static int parse_int(char **buffer,size_t *len)
{
	int X=0;
	if (*len>=sizeof(int)) {
		memcpy((void*) &X, *buffer, sizeof(int));
		*buffer+= sizeof(int);
		*len-= sizeof(int);
	}
	return X;
}

static char* parse_str(char **buffer,size_t *len) {
	size_t L;
	L= parse_int(buffer,len);
	if (L && *len>=L) {
		char *X;
		X= (char*) malloc(L+1);
		memcpy(X, *buffer, L);
		X[L]= 0;
		*buffer+= L;
		*len-= L;
		return X;
	} else return 0;
}



void $CLASS::boxed_int::deserialize(char **buffer,size_t *len)
{
	value= parse_int(buffer,len);
	if (value || table->show_if_zero(row,col)) {
		char *X;
		X= table->format_int(value, 
			table->cell_defs[row][col].format);
		table->set_cell_text_noupdate(row,col,X);
		free(X);
	} else table->set_cell_text_noupdate(row,col,0);
}

void $CLASS::boxed_str::deserialize(char **buffer,size_t *len)
{
	if (value) free(value);
	value= parse_str(buffer,len);
	table->set_cell_text_noupdate(row,col,value);
}


void $CLASS::deserialize_data(char *buf,size_t len)
{
	char **B;
	size_t *L;
	B= &buf;
	L= &len;

	$DESERIALIZATION
	calculate_cell_content_geometries();
	calculate_col_widths();
	calculate_row_heights();
	convert_to_actual_coordinates();
	layout();
	update();
}

void $CLASS::detach()
	{ FXScrollArea::detach();}

int $CLASS::getContentWidth()
	{  return x_offsets[$NCOLS]; }

int $CLASS::getContentHeight()
	{ return y_offsets[$NROWS]; }

int $CLASS::getDefaultWidth()
	{ return x_offsets[$NCOLS]+vertical->getDefaultWidth() + 10;}

int $CLASS::getDefaultHeight()
	{ return y_offsets[$NROWS]+horizontal->getDefaultHeight()+5; }

void $CLASS::layout() {
	FXScrollArea::layout();
#if 0
	int K;
	K= x_offsets[$NCOLS];
	if (K<width) x_pad= (width-K)/2;
	K= y_offsets[$NROWS];
	if (K<height) y_pad= (height-K)/2;
#endif
}


long $CLASS::onPaint(FXObject* ob,FXSelector sel,void* ptr){
	FXEvent* event;
	crect_t R;
	int i,j;

	event=(FXEvent*)ptr;

	FXDCWindow dc(this,event);

	/* FIXME: subtract the painted areas from
		this and then do the paint */
	dc.setForeground(getBackColor());
	dc.fillRectangle(event->rect.x,event->rect.y,
		event->rect.w, event->rect.h);

	R.x1= event->rect.x- pos_x;
	R.y1= event->rect.y- pos_y;
	R.x2= R.x1+event->rect.w;
	R.y2= R.y1+event->rect.h;

	/* now, R is in document coordinates */

/*	printf("PAINT! %d,%d  [%d,%d]\n",R.x1,R.y1,R.x2,R.y2);
	printf("WINDOW: %d %d\n",getWidth(), getHeight());*/

	R.x1-= x_pad;
	R.x2-= x_pad;
	R.y1-= y_pad;
	R.y2-= y_pad;

	for(i=0;i<$NBGFILLS;i++)
		if (rectangles_intersect(R, bgfills[i].area)) {
			dc.setForeground(bgfills[i].color);
			dc.fillRectangle(
				x_pad+bgfills[i].area.x1+pos_x,
				y_pad+bgfills[i].area.y1+pos_y,
				bgfills[i].area.x2-bgfills[i].area.x1,
				bgfills[i].area.y2-bgfills[i].area.y1);
		}

	dc.setForeground(rgb24tofxcolor(0x0));
	for(i=0;i<$NHORZLINES;i++)
		if (xline_visible(R, horz_lines[i]))
			dc.drawLine(
				x_pad+horz_lines[i].x1+pos_x,
				y_pad+horz_lines[i].y+pos_y,
				x_pad+horz_lines[i].x2+pos_x,
				y_pad+horz_lines[i].y+pos_y);

	for(i=0;i<$NVERTLINES;i++)
		if (yline_visible(R, vert_lines[i]))
			dc.drawLine(
				x_pad+vert_lines[i].x+pos_x,
				y_pad+vert_lines[i].y1+pos_y,
				x_pad+vert_lines[i].x+pos_x,
				y_pad+vert_lines[i].y2+pos_y);

	for(i=0;i<$NROWS;i++)
		for(j=0;j<$NCOLS;j++) {
			cell_t *C;
			const cell_def_t *def;
			crect_t TR;
			C= &cells[i][j];
			def= &cell_defs[i][j];
			if (def->kind!=ck_header && def->kind!=ck_data)
				continue;
			if (!C->data || !C->data[0])
				continue;
			TR.x1= x_offsets[j];
			TR.y1= y_offsets[i];
			TR.x2= x_offsets[j+def->colspan];
			TR.y2= y_offsets[i+def->rowspan];
			if (!rectangles_intersect(R, TR))
				continue;
			render_text(&dc, TR, def->valign, def->halign,
				def->kind==ck_header, C->data,
				C->cwidth, C->cheight);
		}
	
	return 1;
}

long $CLASS::onClick(FXObject* ob,FXSelector sel,void* ptr) {
	FXEvent* event;
	int x,y;
	int r,c;

	event=(FXEvent*)ptr;

	if (!editable) return 1;
	if (editor) {
		delete editor;
		editor= 0;
		return 1;
	}

	handle(this,FXSEL(SEL_FOCUS_SELF,0),ptr);
	x= event->win_x-pos_x; y= event->win_y-pos_y;

	y-= y_pad;
	x-= x_pad;

	for(r=0;r<$NROWS;r++)
		if (y>=y_offsets[r] && y<y_offsets[r+1])
			break;
	if (r==$NROWS) return 1;

	for(c=0;c<$NCOLS;c++)
		if (x>=x_offsets[c]&&x<x_offsets[c+1])
			break;
	if (c==$NCOLS) return 1;

	if (cell_defs[r][c].kind==ck_merged) {
		int a,b;
		a= r; b=c;
		r= cell_defs[a][b].merge_r;
		c= cell_defs[a][b].merge_c;
	}
	if (cell_defs[r][c].kind==ck_empty) return 1;
	if (cell_defs[r][c].kind==ck_header) return 1;
	if (!cell_defs[r][c].editable) return 1;

	start_edit(r,c);
	return 1;
}

void $CLASS::make_cell_visible(int r,int c)
{
	int X= x_offsets[c]+pos_x;
	int Y= y_offsets[r]+pos_y;
	int W= x_offsets[c+cell_defs[r][c].colspan]-x_offsets[c];
	int H= y_offsets[r+cell_defs[r][c].rowspan]-y_offsets[r];
	int dx,dy;

	dx= dy= 0;

	if (X<0) { dx= X; X= 0; }

	if (X+W>=viewport_w) {
		if (X!=0) {
			dx+= -X;
			X= 0;
		}
	}

	if (Y<0) { dy= Y; Y= 0; }

	if (Y+H>=viewport_h) {
		if (Y!=0) {
			dy+= -Y;
			Y= 0;
		}
	}

	if (dx || dy) 
		setPosition(pos_x+dx,pos_y+dy);
}

void $CLASS::start_edit(int r,int c) {

	if (r<0 || c<0 || r>=$NROWS || c>=$NCOLS)
		return ;
	make_cell_visible(r,c);
	editor= new FXTextField(this, 1, NULL, 0,
		FRAME_THICK, 
			x_pad+x_offsets[c]+pos_x,
			y_pad+y_offsets[r]+pos_y,
			x_offsets[c+cell_defs[r][c].colspan]-x_offsets[c],
			y_offsets[r+cell_defs[r][c].rowspan]-y_offsets[r]);
	editor->create();
	if (cells[r][c].data)
		editor->setText(cells[r][c].data);
	editor->setFocus();
	editor->setTarget(this);
	editor->setSelector(ID_UPDATECELL);
	editor->show();
	input_r= r;
	input_c= c;
}

long $CLASS::onCellChange(FXObject* ob,FXSelector sel,void* ptr) {
	char *text;
	int intv;
	char *strv;
	int idx1,idx2;
	text= strdup(editor->getText().text());
	delete editor;
	editor= 0;

	strv= 0;
	switch (cell_defs[input_r][input_c].base_type) {
	case cell_int:
		intv= unformat_int(text,
				cell_defs[input_r][input_c].format);
		break;
	case cell_str:
		strv= unformat_str(text,
				cell_defs[input_r][input_c].format);
	}
	idx1= cell_defs[input_r][input_c].idx1;
	idx2= cell_defs[input_r][input_c].idx2;
	switch(cell_defs[input_r][input_c].skind) {
	$CELL_CHANGE_CALLBACK_CASES
	}
	free(text);

	if (strv) free(strv);
	return 1;
}

int $CLASS::col(const char* X)
{
	int S=0;
	while(*X) {
		S= S*26 + X[0]-'A';
		X++;
	}
	return S;
}

int $CLASS::fit_to_range(int x,int min,int max) {
	if (x<min) return min;
	else if (x>max) return max;
	else return x;
}

static void freeifnnull(char *x) {
	if (x) free(x);
}

$CLASS::$CLASS() { flags|=FLAG_ENABLED|FLAG_SHOWN;}
$CLASS::~$CLASS() {
	int i,j;
	for(i=0;i<$NROWS;i++)
		for(j=0;j<$NCOLS;j++)
			if (cells[i][j].data)
				free(cells[i][j].data);

	$STORAGE_DESTRUCTOR_CODE
}

$CLASS::$CLASS(FXComposite* p,
		FXuint opts,
		FXint x,FXint y,FXint w,FXint h)
	: FXScrollArea(p,opts,x,y,w,h)
{
	int i,j;
	flags |= FLAG_ENABLED;
	myapp= FXApp::instance();
	for(i=0;i<$NROWS;i++)
		for(j=0;j<$NCOLS;j++) {
			cells[i][j].data= 0;
			cells[i][j].cwidth= cells[i][j].cheight= 0;
		}
	$CELL_INIT_CODE

	for(i=0;i<$NROWS;i++) row_heights[i]= 0;
	for(i=0;i<$NCOLS;i++) col_widths[i]= 0;
	editor= 0;

	$STORAGE_INIT_CODE

	x_pad= 0;
	y_pad= 0;
}

void $CLASS::convert_to_actual_coordinates() {
	int i;
	int S;
	
	for(S=0,i=0;i<=$NCOLS;i++) {
		x_offsets[i]= S;
		S+= col_widths[i];
	}

	for(S=0,i=0;i<=$NROWS;i++) {
		y_offsets[i]= S;
		S+= row_heights[i];
	}

	for(i=0;i<$NHORZLINES;i++) {
		horz_lines[i].y= y_offsets[horz_line_defs[i].y];
		horz_lines[i].x1= x_offsets[horz_line_defs[i].x1];
		horz_lines[i].x2= x_offsets[horz_line_defs[i].x2];
	}
	for(i=0;i<$NVERTLINES;i++) {
		vert_lines[i].x= x_offsets[vert_line_defs[i].x];
		vert_lines[i].y1= y_offsets[vert_line_defs[i].y1];
		vert_lines[i].y2= y_offsets[vert_line_defs[i].y2];
	}
	for(i=0;i<$NBGFILLS;i++) {
		const bgfill_def_t *D;
		bgfill_act_t *A;
		D= bgfill_defs + i;
		A= bgfills+i;
		A->color= rgb24tofxcolor(D->color);
		A->area.x1= x_offsets[D->area.x1];
		A->area.y1= y_offsets[D->area.y1];
		A->area.x2= x_offsets[D->area.x2];
		A->area.y2= y_offsets[D->area.y2];
	}
}


FXColor $CLASS::rgb24tofxcolor(unsigned int x) {
	unsigned int r,g,b;
	r= (x>>16)&0xff;
	g= (x>>8) & 0xff;
	b= (x)&0xff;
	return FXRGB(r,g,b);
}

void $CLASS::get_text_size(int bold,char *data, int *w,int *h)
{
	FXFont *font;
	char *x;
	int W=0;
	int H= 0;
	int tw,th;

	*w=0; *h= 0;
	font= bold ? header_font : data_font;
	if (!data || !data[0]) return ;
	
	while((x=strchr(data,'\n'))) {
		x[0]= 0;
		get_single_line_text_size(font, data, &tw, &th);
		x[0]= '\n';
		H+= th;
		if (tw>W) W= tw;
		data= x+1;
	}

	if (*data) {
		get_single_line_text_size(font, data, &tw, &th);
		H+= th;
		if (tw>W) W= tw;
	}

	*w= W;
	*h= H;
}

void $CLASS::get_single_line_text_size
	(FXFont* font,char*data,int *tw,int *th)
{
	*tw= font->getTextWidth(data);
	*th= font->getFontHeight();
}

void $CLASS::render_text
	(FXDC* dc,crect_t area, int valign,int halign, int bold,
	char *data,int content_width,int content_height)
{
	int X,Y;
	char *L;
	FXFont* font;

	if (!data || !data[0]) return ;
	font= bold ? header_font : data_font;

	X= area.x1;
	Y= area.y1;
	switch(valign) {
		case valign_top: break;
		case valign_bottom:
			Y= area.y2- content_height;
			break;
		case valign_center:
			Y= (area.y1+area.y2-content_height)/2;
			break;
	}
	while((L=strchr(data,'\n'))) {
		L[0]= 0;
		Y= render_text_line(dc,font,X, Y,
				halign, data,area.x2-area.x1);
		data= L+1;
		L[0]= '\n';
	}
	if (data[0])
		render_text_line(dc,font,X, Y,
				halign, data,area.x2-area.x1);
}



int $CLASS::render_text_line
		(FXDC* dc,FXFont *font,int X, int Y,
		int halign, char *data, int avail_w)
{
	switch(halign) {
	case halign_left: break;
	case halign_right:
		X= X+avail_w-font->getTextWidth(data); 
		break;
	case halign_center:
		X= X+avail_w/2-font->getTextWidth(data)/2;
		break;
	}
	dc->setFont(font);
	dc->drawText(x_pad+X+pos_x,
		y_pad+Y+font->getFontAscent()+pos_y,
		data, strlen(data));
	return Y+font->getFontHeight();
}

void $CLASS::calculate_cell_geometry(int i,int j) {
	switch(cell_defs[i][j].kind) {
	case ck_header: 
		get_text_size(1, cells[i][j].data,
		&cells[i][j].cwidth, &cells[i][j].cheight);
		break;
	case ck_data:
		if (cells[i][j].data && cells[i][j].data[0]) {
			get_text_size(0, cells[i][j].data,
			&cells[i][j].cwidth, &cells[i][j].cheight);
			break;
		}
	default:
		cells[i][j].cwidth= 0;
		cells[i][j].cheight= 0;
	}
}

void $CLASS::calculate_cell_content_geometries()
{
	int i,j;
	for(i=0;i<$NROWS;i++)
		for(j=0;j<$NCOLS;j++)
			calculate_cell_geometry(i,j);
}

int $CLASS::calculate_row_heights()
{
	int i,j;
	int nh[$NROWS];
	for(i=0;i<$NROWS;i++) 
		nh[i]= 0;
	for(i=0;i<$NROWS;i++) 
		for(j=0;j<$NCOLS;j++)  {
			int k;
			int a;
			if (!cell_defs[i][j].rowspan) continue;
			k= ceil(cells[i][j].cheight,cell_defs[i][j].rowspan);
			for(a=i;a<i+cell_defs[i][j].rowspan;a++)
				if (k>nh[a]) nh[a]= k;
		}
	for(i=0;i<$NROWS;i++)
		if (nh[i]!=row_heights[i]) {
			memcpy(row_heights, nh, sizeof(int)*$NROWS);
			return 1;
		}
	return 0;
}

int $CLASS::calculate_col_widths()
{
	int i,j;
	int nw[$NCOLS];
	for(j=0;j<$NCOLS;j++)  
		nw[j]= 0;

	for(j=0;j<$NCOLS;j++) 
		for(i=0;i<$NROWS;i++)  {
			int k;
			int a;
			if (!cell_defs[i][j].colspan) continue;
			k= ceil(cells[i][j].cwidth,cell_defs[i][j].colspan);
			for(a=j;a<j+cell_defs[i][j].colspan;a++)
				if (k>nw[a]) nw[a]= k;
		}
	for(j=0;j<$NCOLS;j++)  
		if (nw[j]!=col_widths[j]) {
			memcpy(col_widths, nw, sizeof(int)*$NCOLS);
			return 1;
		}
	return 0;
}

int $CLASS::ceil(int a,int b) { return (a+b-1)/b; }

int $CLASS::classify_point(crect_t R, int x,int y)
{
	int C;
	C= 0;
	if (x<R.x1) C|= 1; else if (x>=R.x2) C|=2;
	if (y<R.y1) C|=4; else if (y>=R.y2) C|= 8;
	return C;
}

int $CLASS::rectangles_intersect(crect_t R1,crect_t R2)
{
	return (classify_point(R1, R2.x1,R2.y1) &
		 classify_point(R1, R2.x2,R2.y2))==0;
}

int $CLASS::xline_visible(crect_t R, xline_t L)
{
	return (classify_point(R, L.x1, L.y) &
		 classify_point(R, L.x2, L.y))==0;
}

int $CLASS::yline_visible(crect_t R,yline_t L)
{
	return (classify_point(R, L.x, L.y1) &
		 classify_point(R, L.x, L.y2))==0;
}

const $CLASS::xline_t
$CLASS::horz_line_defs[$NHORZLINES] = {
	$HORZ_LINE_DEFS
};
const $CLASS::yline_t
$CLASS::vert_line_defs[$NVERTLINES] = {
	$VERT_LINE_DEFS
};

const $CLASS::bgfill_def_t
$CLASS::bgfill_defs[$NBGFILLS] = {
	$BG_FILL_DEFS
};

const $CLASS::cell_def_t 
$CLASS::cell_defs[$NROWS][$NCOLS]= {
	$CELL_DEFS
};


void $CLASS::set_cell_text_noupdate(int r,int c,char *text) {
	if (c<0 || r<0 || c>=$NCOLS || r>=$NROWS) return ;
	if (cell_defs[r][c].kind!=ck_data) return ;
	if (cells[r][c].data)
		free(cells[r][c].data);
	if (text) text= strdup(text);
	cells[r][c].data= text;
}

void $CLASS::set_cell_text(int r,int c,char *text) {
	if (c<0 || r<0 || c>=$NCOLS || r>=$NROWS) return ;
	if (cell_defs[r][c].kind!=ck_data) return ;
	set_cell_text_noupdate(r,c,text);
	calculate_cell_geometry(r,c);
}


void $CLASS::begin_update() {
}

void $CLASS::end_update() {
	if (calculate_col_widths())  {
		convert_to_actual_coordinates();
		recalc();
		layout();
	}
	update();
}


void $CLASS::boxed_str::init($CLASS *tb,int r,int c) {
	value= 0;
	table= tb;
	row= r;
	col= c;
}

void $CLASS::boxed_str::start_edit() {
	table->start_edit(row,col);
}


void $CLASS::boxed_int::start_edit() {
	table->start_edit(row,col);
}

void $CLASS::boxed_int::init($CLASS *tb,int r,int c) {
	value= 0;
	table= tb;
	row= r;
	col= c;
}

void $CLASS::boxed_int::set(int v) {
	char *tmp;
	value = v;
	if (value || table->show_if_zero(row,col)) {
		tmp= table->format_int(v,
			table->cell_defs[row][col].format);
		table->set_cell_text(row,col,tmp);
		free(tmp);
	} else table->set_cell_text(row,col,0);
}

void $CLASS::boxed_str::set(char *v) {
	if (value)
		free(value);
	if (v && v[0]) 
		value= strdup(v);
	else
		value= 0;
	table->set_cell_text(row,col,value);
}

char* $CLASS::format_int(int v,const char *fmt)
{
	size_t ndigits;
	char buf[200];
	int i;
	char *D;
	char *x;
	char *b;
	size_t L;

	sprintf(buf,"%d",v);
	if (!fmt || !fmt[0])
		return strdup(buf);

	ndigits= 0;
	for(i=0;fmt[i];i++)
		if (fmt[i]=='#') ndigits++;

	L=  (ndigits>strlen(buf) ? ndigits : strlen(buf)) +
		strlen(fmt)-ndigits + 1;
	D= (char*) malloc(L);
	x= D+L-1;
	x[0]= 0;
	x--;
	b= buf+strlen(buf)-1;

	for(i=strlen(fmt)-1;i>=0;i--) {
		if (fmt[i]=='#') {
			if (b<buf) *x= '0';
			else {
				*x= *b;
				b--;
			}
			x--;
		} else {
			*x= fmt[i];
			x--;
		}
	}

	while(b>=buf) {
		*x= *b;
		x--; b--;
	}
	return D;
}

char *$CLASS::format_str(char *v,const char* fmt)
{
	if (v)
		return strdup(v);
	else
		return 0;
}

char *$CLASS::unformat_str(char *v,const char* fmt)
{
	if (v)
		return strdup(v);
	else
		return 0;
}

int $CLASS::unformat_int(char *x,const char*fmt)
{
	int i;
	int j;
	if (!x) return 0;
	x= strdup(x);
	j= 0;
	for(i=0;x[i];i++)
		if ((x[i]>='0' && x[i]<='9') || (j==0 && x[i]=='-'))
			x[j++]= x[i];
	x[j]= 0;
	i= atoi(x);
	free(x);
	return i;
}

$CLASS::raw_storage_t* $CLASS::parse_data
	(char *buf,size_t len)
{
	char **B;
	size_t *L;
	raw_storage_t* obj;
	obj= (raw_storage_t*) malloc(sizeof(*obj));
	B= &buf;
	L= &len;

	$RAW_PARSING_CODE
	return obj;
}

void $CLASS::destroy_data($CLASS::raw_storage_t *obj) {
	$RAW_DESTRUCTOR
	free(obj);
}

-[end]
0};


=[quote code, begin ]
![bin2c]
=[quote code, end ]
![end]

=[]

-[public] void print_code
	(char *sn,strpairlist_t *subs)
{
	char buf[2000];
	sprintf(buf,"%s.cc", sn);
	print_code_with_subs(buf,source_code, subs);
	sprintf(buf,"%s.h", sn);
	print_code_with_subs(buf,header_code, subs);
}

-[private] void print_code_with_subs
	(char *fn, char *code, strpairlist_t *subs)
{
	FILE *f;
	f= fopen(fn,"w");
	if (!f)
		exit(fprintf(stderr,"can not open %s for writing: %s\n",
			fn, strerror(errno)));
	code= do_replacements(code, subs);
	fwrite(code, 1, strlen(code), f);
	fclose(f);
}


-[ headers, begin ]
	!stdlib !stdio !string !errno repstr
-[end]


