-[ c library "decl.c", header "decl.h" ]

-[ headers,begin] !stdlib !stdio !string
parsedoc style miscfn -[end]

-[public] enum celltype { typ_int, typ_str };

-[public] typedef struct {
	char *name;
	int editable;
	enum celltype type;
	char *format;
	int show_if_zero;
} decl_t;

-[declist_t : public list of decl_t* ]

-[public] declist_t* parse_declarations(grid_t* G)
{
	declist_t *L;
	int i,j;
	L= declist_new();
	for(i=0;i<G->nrows;i++) 
		for(j=0;j<G->ncols;j++)
			if (G->cells[i][j].kind==ck_data) {
				parse_declaration(L,G,i,j);
				j= G->ncols;
			}
	return L;

}


-[private] void parse_declaration
	(declist_t* L,grid_t *G,int i,int j)
{
	decl_t *D;
	char *C;
	-[ allocate object D ]
	D->name= get_cell(G,i,j);
	if (!D->name)
		exit(fprintf(stderr,"empty name in "
			"declaration at row %d\n", i+1));
	j++;
	C= get_cell(G,i,j);
	
	if (!C) 
		exit(fprintf(stderr,"empty type in "
			"declaration at row %d\n", i+1));
	-[  string switch C, begin ]
		-[case "int"] D->type= typ_int; break;
		-[case "str"] D->type= typ_str; break;
		default:
			exit(fprintf(stderr,"unknown type in "
				"declaration at row %d : %s\n", i+1,C));
	-[ end ]
	j++;
	D->format= get_cell(G,i,j);
	j++;
	C= get_cell(G,i,j);
	if (!C)
		exit(fprintf(stderr,"empty editable field in "
			"declaration at row %d\n", i+1));
	-[ string switch C, begin ]
		-[case "y"]
		-[case "Y"] D->editable= 1; break;
		-[case "n"]
		-[case "N"] D->editable=0; break;
		default:
		exit(fprintf(stderr,"unknown editable field in "
			"declaration at row %d\n", i+1));
	-[ end ]
	j++;
	C= get_cell(G,i,j);
	D->show_if_zero= 0;
	if (C && D->type==typ_int) {
		-[ string switch C, begin ]
			-[case "y"]
			-[case "Y"] D->show_if_zero= 1; break;
		-[ end ]
	}
	declist_append(L,D);
}

-[private] char* get_cell(grid_t *G,int r,int c)
{
	char *data;
	if (r<0 || r>=G->nrows) return 0;
	if (c<0 || c>=G->ncols) return 0;
	if (!G->cells[r][c].data)
		return 0;
	data= strdup(G->cells[r][c].data);
	remove_spaces(data);	
	return data;
}

