-[ c library "genlines.c", header "genlines.h" ]

-[headers, begin]
	!stdlib !string !stdio
	parsedoc
	repstr style
-[end]


-[private] typedef struct { int a; int b1; int b2; } sline_t;
-[cbuffer_t: private buffer [1000] of char ]
-[sbuffer_t: private buffer [200] of sline_t ]

-[public] void gen_lines(grid_t *G,strpairlist_t *vars)
{
	int i,j;
	sbuffer_t *buffer;

	buffer= sbuffer_new();
	-[ generate xlines ]
	do_lines( buffer, "$NHORZLINES", "$HORZ_LINE_DEFS", vars);

	buffer= sbuffer_new();
	-[ generate ylines ]
	do_lines( buffer, "$NVERTLINES", "$VERT_LINE_DEFS", vars);
}

-[private] void do_lines
	(sbuffer_t *buffer,
	char *name_n,char *name_def, strpairlist_t *vars)
{
	cbuffer_t *obuffer;
	int nlines;
	sline_t *lines;
	char tbuf[20];
	int dummy;

	lines= sbuffer_collectda(buffer, &nlines);
	lines= combine_lines(lines, &nlines);
	sprintf(tbuf,"%d",nlines);
	strpairlist_append(vars,
		strpair_new(strdup(name_n), strdup(tbuf)));

	obuffer= cbuffer_new();
	put_lines(obuffer, lines, nlines);
	free(lines);
	strpairlist_append(vars,
		strpair_new(strdup(name_def),
			cbuffer_collectda(obuffer,&dummy)));
}


-[private] void put_lines(cbuffer_t *buffer,sline_t *lines,int nlines)
{
	int i;
	char xb[1000];
	for(i=0;i<nlines;i++) {
		int j;
		sprintf(xb,"{ %d, %d, %d },\n",
			lines[i].a, lines[i].b1, lines[i].b2);
		for(j=0;xb[j];j++)
			cbuffer_append(buffer, xb[j]);
	}
	cbuffer_append(buffer,0);
}

-[private] sline_t *combine_lines(sline_t *lines, int *n)
{
	int N;

	N= *n;
	-[ quicksort array lines = sline_t[N] , begin ]
		-[compare X1 to X2]
			result= X1->a - X2->a;
			result= (result==0) ? (X1->b1 - X2->b1) : result;
			result= (result==0) ? (X1->b2 - X2->b2) : result;
	-[ end ]

	do {
		*n= N;
		-[ combine neighbouring lines ]
	} while(N !=*n );

	return lines;
}

=[ combine neighbouring lines ]
	-[ unify array lines : N, begin ]
		result= 1;
		if (lines[i].a==lines[j].a) {
			if (lines[i].b1 >= lines[j].b1&&lines[i].b1<lines[j].b2)
			{
				if (lines[j].b2>lines[i].b2)
					lines[i].b2= lines[j].b2;
				lines[i].b1= lines[j].b1;
				result= 0;
			} else if (lines[i].b2>=lines[j].b1
					&& lines[i].b2<lines[j].b2) {
				if (lines[j].b1<lines[i].b1)
					lines[i].b1= lines[j].b1;
				lines[i].b2= lines[j].b2;
				result= 0;
			}
		}
	-[ end ]
	

=[ generate xlines ]
for(i=0;i<G->nrows;i++)
	for(j=0;j<G->ncols;j++) {
		xcell_t *X;
		X= &G->cells[i][j];
		if (X->kind==ck_merged)
			continue;
		if (!X->colspan || !X->rowspan)
			continue;
		if (X->border&b_bottom) {
			sline_t L;
			L.a= i+X->rowspan;
			L.b1= j;
			L.b2= j+X->colspan;
			sbuffer_append(buffer,L);
		}
		if (X->border&b_top) {
			sline_t L;
			L.a= i;
			L.b1= j;
			L.b2= j+X->colspan;
			sbuffer_append(buffer,L);
		}
}

=[ generate ylines ]
for(j=0;j<G->ncols;j++) 
	for(i=0;i<G->nrows;i++) {
		xcell_t *X;
		X= &G->cells[i][j];
		if (X->kind==ck_merged)
			continue;
		if (!X->colspan || !X->rowspan)
			continue;
		if (X->border&b_left) {
			sline_t L;
			L.a= j;
			L.b1= i;
			L.b2= i+X->rowspan;
			sbuffer_append(buffer,L);
		}
		if (X->border&b_right) {
			sline_t L;
			L.a= j+1;
			L.b1= i;
			L.b2= i+X->rowspan;
			sbuffer_append(buffer,L);
		}
}

=[]



-[ unused symbol cbuffer_collectd ]
-[ unused symbol sbuffer_collectd ]

