-[ c library "parsedoc.c", header "parsedoc.h" ]


-[ headers, begin ]
	!stdio !expat
-[ end ]

-[public] typedef struct {
	int row,col;
	char *data;
} cell_t;

-[public] typedef struct {
	char *data;
} merge_t;

-[public] typedef struct {
	int startCol, startRow, endCol, endRow;
	int hAlign, vAlign, bgcolor, fgcolor;
	int bTop, bLeft, bRight, bBottom;
	int bold;
} styleregion_t;

-[ cellist_t : public list of cell_t* ]
-[ mergelist_t : public list of merge_t* ]
-[ styleregionlist_t : public list of styleregion_t* ]

-[public] typedef struct {
	char *name;
	char *maxcol, *maxrow;
	cellist_t *cells;
	mergelist_t *merges;
	styleregionlist_t *styleregions;
} sheet_t;

-[ sheetlist_t : public list of sheet_t* ]

-[public] typedef struct {
	sheetlist_t *sheets;
} doc_t;



-[public] doc_t *parse_doc(FILE *f)
{
	doc_t *doc;
-[ P= xml parser, begin ]
-[pnode "gnm:Workbook" ,begin] return 0;
-[pnode "gnm:Sheets", begin ] return 0;
-[pnode "gnm:Sheet", begin ]
	doc_t *D= udata;
	-[ private ] sheet_t *csheet;
	-[ allocate object csheet ]
	csheet->cells= cellist_new();
	csheet->styleregions= styleregionlist_new();
	csheet->merges= mergelist_new();
	sheetlist_append( D->sheets, csheet);
	return 0;
	-[ pnode "gnm:Name", begin ]
		return &csheet->name;
	-[ end ]
	-[ pnode "gnm:MaxRow", begin ]
		return &csheet->maxrow;
	-[ end ]
	-[ pnode "gnm:MaxCol", begin ]
		return &csheet->maxcol;
	-[ end ]
	-[ pnode "gnm:Styles", begin ] return 0;
	-[ pnode "gnm:StyleRegion", begin ]
		styleregion_t *S;
		-[ allocate object S ]
		styleregionlist_append(csheet->styleregions, S);
		-[ private ] styleregion_t *cstyle;
		cstyle= S;
		S->startCol= get_int_attrib(attr,"startCol");
		S->startRow= get_int_attrib(attr, "startRow");
		S->endCol= get_int_attrib(attr,"endCol");
		S->endRow= get_int_attrib(attr, "endRow");
		return 0;
		-[ pnode "gnm:Style", begin ]
			cstyle->hAlign= get_int_attrib(attr,"HAlign");
			cstyle->vAlign= get_int_attrib(attr,"VAlign");
			cstyle->fgcolor= get_color_attrib(attr,"Fore");
			cstyle->bgcolor= get_color_attrib(attr,"Back");
			return 0;
			-[ pnode "gnm:Font" , begin ]
				cstyle->bold= get_int_attrib(attr,"Bold");
				return 0;
			-[ end ]
			-[ pnode "gnm:StyleBorder", begin ] return 0;
				-[ pnode "gnm:Top", begin ]
					cstyle->bTop= 1;
					return 0;
				-[ end ]
				-[ pnode "gnm:Left", begin ]
					cstyle->bLeft= 1;
					return 0;
				-[ end ]
				-[ pnode "gnm:Right", begin ]
					cstyle->bRight= 1;
					return 0;
				-[ end ]
				-[ pnode "gnm:Bottom", begin ]
					cstyle->bBottom= 1;
					return 0;
				-[ end ]
			-[ end ]
		-[ end ]
	-[ end ]
	-[ end ]
	-[ pnode "gnm:Cells", begin ] return 0;
	-[ pnode "gnm:Cell", begin ]
		cell_t *C;
		-[ allocate object C ]
		C->row= get_int_attrib(attr,"Row");
		C->col= get_int_attrib(attr,"Col");
		cellist_append(csheet->cells,C);
		return &C->data;
	-[ end ]
	-[ end ]
	-[ pnode "gnm:MergedRegions", begin ] return 0;
	-[ pnode "gnm:Merge", begin ]
		merge_t *M;
		-[ allocate object M ]
		mergelist_append(csheet->merges, M);
		return &M->data;
	-[ end ]
	-[ end ]
-[end]
-[end] 
-[end]
-[end]

	-[ allocate object doc ]
	doc->sheets= sheetlist_new();

#if 0
	printf("PARSE TREE\n");
	pnode_print(P);
	printf("END PARSE TREE\n");
#endif

	parse_xml(f, P, doc);
	return doc;
}


-[ import XML parser routines in private scope ]



-[private] int get_int_attrib(char **attr,char *name)
{
	char *A;
	A= get_str_attrib(attr,name);
	if (A) return atoi(A);
	return 0;
}

-[private] int get_color_attrib(char **attr,char *name)
{
	char *A;
	char **components;
	int i;
	int icom[3];
	int C;
	A= get_str_attrib(attr,name);
	if (!A) return 0;
	components= strsplit(":",A);
	if (!components) return 0;
	for(i=0;i<3;i++) icom[i]= 0;
	for(i=0;i<3;i++)
		if (components[i]) {
			icom[i]= phex(components[i]);
			free(components[i]);
		}
	free(components);
	C= 0;
	for(i=0;i<3;i++)
		C= (C<<8) | (icom[i]>>8);
	return C;
}

-[private] int phex(char *x)
{
	int A;
	int i;
	A= 0;
	for(i=0;i<4 && x[i];i++) {
		int Z;
		if (x[i]>='0' && x[i]<='9') Z = x[i]-'0';
		else if  (x[i]>='A' && x[i]<='F')  Z= x[i]-'A'+10;
		else if  (x[i]>='a' && x[i]<='f')  Z= x[i]-'a'+10;
		else Z= 0;
		A= (A<<4) | Z;
	}
	return A;
}


-[private] char* get_str_attrib(char **attr,char *name) {
	int L;
	int i;
	for(L=0;attr[L];L++)
		;
	for(i=0;i<L;i+=2)
		if (!strcmp(attr[i],name))
			return attr[i+1];
	return 0;
}

-[ import strsplit function in private scope ]

