-[ c library "part.c", header "part.h" ]

-[ headers, begin ]
	!stdio !stdlib !string
	parsedoc style repstr
	miscfn
-[end]

-[public] void partition_bgfills(grid_t *G,strpairlist_t *vars)
{
	int i;
	int N;
	int ncolors;
	int *colors;
	char buf[20];
	cbuffer_t *obuffer;

	N= 0;

	obuffer= cbuffer_new();

	colors= get_colors(G, &ncolors);
	for(i=0;i<ncolors;i++)
		N+= partition_bgfills_with_color(colors[i],G,obuffer);
	free(colors);

	sprintf(buf,"%d", N);
	strpairlist_append(vars, 
		strpair_new(strdup("$NBGFILLS"),strdup(buf)));

	cbuffer_append(obuffer,0);

	strpairlist_append(vars,
		strpair_new(strdup("$BG_FILL_DEFS"),
			cbuffer_collectda(obuffer, &N)));
}

-[private] int* get_colors(grid_t *G,int *n) {
	int i,j;
	int *colors;
	int ncolors;

	-[ colors= allocate array G->nrows*G->ncols]
	ncolors= 0;

	for(i=0;i<G->nrows;i++)
		for(j=0;j<G->ncols;j++) {
			if (G->cells[i][j].kind==ck_merged)
				continue;
			colors[ncolors++]= G->cells[i][j].bgcolor;
		}

	-[quicksort array colors = int [ncolors] , begin]
		-[compare X1 to X2]  result= X1[0]-X2[0];
	-[end]
	-[ unify array colors : ncolors, begin ]
		result= colors[i]-colors[j];
	-[ end ]

	*n= ncolors;
	colors= realloc(colors, ncolors*sizeof(int));
	return colors;
}

-[private] int partition_bgfills_with_color
	(int color,grid_t *G,cbuffer_t *out)
{
	int np;
	image_t *M;
	int i,j;
	M= image_new(G->ncols, G->nrows);
	for(i=0;i<G->nrows;i++) 
		for(j=0;j<G->ncols;j++) {
			xcell_t *X;
			int a,b;
			X= &G->cells[i][j];
			if (X->kind==ck_merged ||
			    X->bgcolor!=color)
				continue;
			for(a=0;a<X->rowspan;a++)
				for(b=0;b<X->colspan;b++) 
					M->data[i+a][j+b]= 1;
		}
	np= partition_image(color,M,out);
	free_image(M);
	return np;
}

-[private] int partition_image(int color, image_t *M,cbuffer_t *out)
{
	int np;
	np= 0;
	while(!image_empty(M)) {
		image_t *C;
		C= get_connected_part(M);
		subtract_image(M,C);
		np+= partition_connected_part(color,C, out);
		free_image(C);
	}
	return np;
}

-[private] image_t *get_connected_part(image_t *G)
{
	image_t *C;
	int i,j;
	C= image_new(G->width,G->height);
	for(i=0;i<G->height;i++)
		for(j=0;j<G->width;j++)
			if (G->data[i][j]) {
				C->data[i][j]= 1;
				search_connected(G,C,i,j);
				return C;
			}
	return 0;
}

-[private] void search_connected(image_t *G,image_t *C,int y,int x)
{
	int nx,ny;
	int K;
	for(K=0;K<4;K++) 
		if (get_neighbour(G,x,y,K,&nx,&ny) &&
			G->data[ny][nx] && !C->data[ny][nx])
		{
			C->data[ny][nx]= 1;
			search_connected(G,C,ny,nx);
		}
}

-[private] int get_neighbour
	(image_t *G,int x,int y,int K,int *nx,int *ny)
{
	switch(K) {
	case 0: *nx= x-1; *ny= y; return *nx>=0; 
	case 1: *nx= x+1; *ny= y; return *nx<G->width;
	case 2: *nx= x; *ny= y-1; return *ny>=0;
	case 3: *nx= x; *ny= y+1; return *ny<G->height;
	}
	return 0;
}

-[private] void subtract_image(image_t *M,image_t *C)
{
	int i,j;
	for(j=0;j<M->height;j++)
	for(i=0;i<M->width;i++)
		M->data[j][i] ^= C->data[j][i];
}

-[private] image_t *image_new(int W,int H)
{
	int i;
	image_t *K;
	-[ allocate object K ]
	K->width= W;
	K->height= H;
	-[ K->data = allocate array H ]
	for(i=0;i<H;i++)
		-[ K->data[i] = allocate array W ]
	return K;
}

-[private] void free_image(image_t *M)
{
	int i;
	for(i=0;i<M->height;i++)
		free(M->data[i]);
	free(M->data);
	free(M);
}

-[private] int image_empty(image_t *M) {
	int i,j;
	for(i=0;i<M->height;i++)
		for(j=0;j<M->width;j++)
			if (M->data[i][j]) return 0;
	return 1;
}

-[private] typedef struct {
	int width,height;
	char **data;
} image_t;


-[private] typedef struct { int x1,y1; int x2,y2; } rect_t;

-[private] int partition_connected_part
	(int color,image_t *M, cbuffer_t *out)
{
	int x,y;
	rect_t V,H;
	char buf[200];
	get_random_point(M, &x,&y);

	V= get_vertical_cut(M,x,y);
	H= get_horizontal_cut(M,x,y);
	if (area(H)<area(V))
		H=V;

	for(x=H.x1;x<H.x2;x++)
		for(y=H.y1;y<H.y2;y++)
			M->data[y][x]= 0;

	sprintf(buf,"{ {%d, %d, %d, %d}, 0x%08x},\n",
		 H.x1, H.y1, H.x2, H.y2,color);
	cbuffer_putstr(out, buf);

	return partition_image(color, M, out) + 1;
}


-[private] void get_random_point(image_t *M,int *X,int *Y)
{
	int i,j;
	int index;
	int Z;
	index= ((double)random()/RAND_MAX)*M->width*M->height;
	Z= 0;
	for(i=0;i<M->height;i++)
		for(j=0;j<M->width;j++) 
			if (M->data[i][j]) {
				int k;
				Z++;
				if (Z==1) { *X= j; *Y=i; }
				for(k=index;k;k/=2)
					if (k==Z) { *X= j; *Y=i; break; }
				if (Z==index) break;
			}
}

-[private] int area(rect_t A) { return (A.x2-A.x1)*(A.y2-A.y1); }


-[private] rect_t get_vertical_cut(image_t *M,int x,int y)
{
	int a;
	rect_t R;
	int sx,sy;
	int ex,ey;
	sx= x;
	for(sx=x;sx;sx--)
		if (!M->data[y][sx-1])
			break;
	for(ex=x+1;ex!=M->width;ex++)
		if (!M->data[y][ex])
			break;
	for(sy=y;sy;sy--) {
		for(a=sx;a<ex;a++)
			if (!M->data[sy-1][a]) goto found1;
	}
found1:
	for(ey=y+1;ey!=M->height;ey++) {
		for(a=sx;a<ex;a++)
			if (!M->data[ey][a]) goto found2;
	}
found2:
	R.x1= sx; R.y1= sy;
	R.x2= ex; R.y2= ey;
	return R;
}


-[private] rect_t get_horizontal_cut(image_t *M,int x,int y)
{
	int a;
	rect_t R;
	int sy,sx;
	int ey,ex;
	sy= y;
	for(sy=y;sy;sy--)
		if (!M->data[sy-1][x])
			break;
	for(ey=y+1;ey!=M->height;ey++)
		if (!M->data[ey][x])
			break;
	for(sx=x;sx;sx--) {
		for(a=sy;a<ey;a++)
			if (!M->data[a][sx-1]) goto found1;
	}
found1:
	for(ex=x+1;ex!=M->width;ex++) {
		for(a=sy;a<ey;a++)
			if (!M->data[a][ex]) goto found2;
	}
found2:
	R.y1= sy; R.x1= sx;
	R.y2= ey; R.x2= ex;
	return R;
}


-[private] void print_image(image_t *M) {
	int y,x;
	printf("-----IMAGE----------\n");
	for(y=0;y<M->height;y++) {
		for(x=0;x<M->width;x++) 
			if (M->data[y][x]) printf("#"); else printf(" ");
		printf("\n");
	}
	printf("-----END----------\n");
}

