-[ c library "serialize.c", header "serialize.h" ]

-[ headers, begin ]
	!stdlib !stdio
	!string
	miscfn repstr
	parsedoc style decl field
-[ end ]

-[public] void output_serialization_code
	(storage_t *S,strpairlist_t *vars)
{
	output_code(vars,S->fields, "$SERIALIZATION", serialize_field);
	output_code(vars,S->fields, "$DESERIALIZATION",
		deserialize_field);
	output_code(vars,S->fields, "$RAW_PARSING_CODE",
		deserialize_raw);
	output_code(vars,S->fields, "$RAW_UNPARSING_CODE",
		serialize_raw);
}

-[private] void output_code
	(strpairlist_t *vars,fieldlist_t *fields, char *vname,
		void (*func)(field_t *,cbuffer_t*))
{
	cbuffer_t *buffer;
	int dummy;
	buffer= cbuffer_new();
	-[list>for F ∈ *fields, begin]
		(*func)(F, buffer);
	-[end]
	cbuffer_append(buffer,0);
	strpairlist_append(vars, strpair_new( strdup(vname),
		cbuffer_collectda(buffer,&dummy)));
}

-[private] void serialize_field(field_t *F,cbuffer_t *B)
{
	call_method_on_field(F,B, "serialize");
}

-[private] void deserialize_field(field_t *F,cbuffer_t *B)
{
	call_method_on_field(F,B, "deserialize");
}

-[private] void call_method_on_field(field_t *F,cbuffer_t *B,char *M) {
	switch(F->kind) {
	case field_single:
		cbprintf(B,"\tS.%s.%s(B,L);\n",F->name,M);
		break;
	case field_array:
	case field_sarray:
		cbprintf(B,"\t{int i;for(i=0;i<%d;i++)\n",F->dim1);
		cbprintf(B,"\t\tS.%s[i].%s(B,L);}\n",F->name,M);
		break;
	case field_2darray:
		cbprintf(B,"\t{int i,j; for(i=0;i<%d;i++)\n",F->dim1);
		cbprintf(B,"\t\tfor(j=0;j<%d;j++)\n",F->dim2);
		cbprintf(B,"\t\t\tS.%s[i][j].%s(B,L);}\n",F->name,M);
		break;
	}
}

-[private] void serialize_raw(field_t *F,cbuffer_t *B)
{
	char *tname=0;
	switch(F->type) {
	case typ_int: tname= "int"; break;
	case typ_str: tname= "str"; break;
	}
	switch(F->kind) {
	case field_single:
		cbprintf(B,"\tunparse_%s(B,L,V->%s);\n",
				tname,F->name);
		break;
	case field_array:
	case field_sarray:
		cbprintf(B,"\t{int i; for(i=0;i<%d;i++)\n",F->dim1);
		cbprintf(B,"\t\tunparse_%s(B,L,V->%s[i]);}\n",
				tname,F->name);
		break;
	case field_2darray:
		cbprintf(B,"\t{int i,j; for(i=0;i<%d;i++)\n",F->dim1);
		cbprintf(B,"\t\tfor(j=0;j<%d;j++)\n",F->dim2);
		cbprintf(B,"\t\t\tunparse_%s(B,L,obj->%s[i][j]);}\n",
				tname,F->name);
		break;
	}
}



-[private] void deserialize_raw(field_t *F,cbuffer_t *B)
{
	char *tname=0;
	switch(F->type) {
	case typ_int: tname= "int"; break;
	case typ_str: tname= "str"; break;
	}
	switch(F->kind) {
	case field_single:
		cbprintf(B,"\tobj->%s=parse_%s(B,L);\n",
				F->name,tname);
		break;
	case field_array:
	case field_sarray:
		cbprintf(B,"\t{int i; for(i=0;i<%d;i++)\n",F->dim1);
		cbprintf(B,"\t\tobj->%s[i]=parse_%s(B,L);}\n",
				F->name,tname);
		break;
	case field_2darray:
		cbprintf(B,"\t{int i,j; for(i=0;i<%d;i++)\n",F->dim1);
		cbprintf(B,"\t\tfor(j=0;j<%d;j++)\n",F->dim2);
		cbprintf(B,"\t\t\tobj->%s[i][j]=parse_%s(B,L);}\n",
				F->name,tname);
		break;
	}
}

