-[ c library "style.c", header "style.h" ]

-[public] enum border { b_left=1, b_right=2, b_top=4, b_bottom=8 };
-[public] enum align { al_begin, al_center, al_end };
-[public] enum cellkind { ck_empty, ck_header, ck_data, ck_merged };

-[public] typedef struct xcell {
	enum cellkind kind;
	int row,col;
	char *data;
	int border;
	int halign, valign;
	int bgcolor, fgcolor;
	int bold;
	int colspan, rowspan;
	struct xcell *merged_to;
} xcell_t;

-[public] typedef struct {
	xcell_t **cells;
	int ncols, nrows;
} grid_t;


-[public] grid_t *apply_styles(sheet_t *sheet)
{
	grid_t *G;
	int i,j;
	-[ allocate object G ]
	G->ncols= 0;
	G->nrows= 0;
	-[list>for C ∈ *sheet->cells, begin]
		if (C->row>G->nrows) G->nrows= C->row;
		if (C->col>G->ncols) G->ncols= C->col;
	-[end]
	G->nrows++;
	G->ncols++;
	-[ G->cells= allocate array G->nrows ]
	for(i=0;i<G->nrows;i++)
		-[ G->cells[i] = allocate array G->ncols ]
	
	for(i=0;i<G->nrows;i++)
		for(j=0;j<G->ncols;j++) {
			G->cells[i][j].row= i;
			G->cells[i][j].col= j;
			G->cells[i][j].kind= ck_empty;
			G->cells[i][j].colspan= 1;
			G->cells[i][j].rowspan= 1;
			G->cells[i][j].bgcolor= 0xffFFff;
		}

	-[list>for K ∈ *sheet->styleregions, begin]
		for(i=K->startRow;i<=K->endRow;i++)
			if (i>=0 && i<G->nrows)
				for(j=K->startCol;j<=K->endCol;j++)
					if (j>=0 && j<G->ncols)
						apply_style(&G->cells[i][j], K);
	-[end]

	-[list>for C ∈ *sheet->cells, begin ]
		xcell_t *X;
		X= &G->cells[C->row][C->col];
		X->data= C->data;
		if (X->data) do_escape_sequences(X->data);
		if (X->bold) 
			X->kind= ck_header;
		else
			X->kind= ck_data;
	-[end]

	-[ list>for R ∈ *sheet->merges, begin ]
		int sR, sC, eR, eC;
		get_merge_coords(R->data,&sR, &sC, &eR, &eC);
		if (sR>=G->nrows || sC>=G->ncols)
			continue;
		if (sR<0) sR= 0;
		if (sC<0) sC= 0;
		if (eR<sR) continue;
		if (eC<sC) continue;
		if (eR==sR && eC==sC) continue;

		G->cells[sR][sC].rowspan= eR-sR+1;
		G->cells[sR][sC].colspan= eC-sC+1;
		for(i=sR;i<=eR;i++)
			for(j=sC;j<=eC;j++) {
				if (i==sR && j==sC)
					continue;
				G->cells[i][j].kind= ck_merged;
				G->cells[sR][sC].border |=
						G->cells[i][j].border;
				G->cells[i][j].merged_to=
					&G->cells[sR][sC];
				G->cells[i][j].rowspan= 0;
				G->cells[i][j].colspan= 0;
			}
	-[end]

	return G;
}


-[private] void apply_style(xcell_t *X, styleregion_t *R)
{
	/* these magic numbers are retrieved by trying
	   gnumeric 1.8.3. I wonder why they didnt provide
	   short strings like center, right, etc. instead of numbers */
	switch(R->hAlign) {
	case 2: X->halign= al_begin; break;
	case 8: X->halign= al_center; break;
	case 4: X->halign= al_end; break;
	}
	switch(R->vAlign) {
	case 1: X->valign= al_begin; break;
	case 4: X->valign= al_center; break;
	case 2: X->valign= al_end; break;
	}
	if (R->bTop) X->border |= b_top;
	if (R->bBottom) X->border |= b_bottom;
	if (R->bLeft) X->border |= b_left;
	if (R->bRight) X->border |= b_right;
	if (R->bold)
		X->bold= 1;
	X->fgcolor= R->fgcolor;
	X->bgcolor= R->bgcolor;
}


-[private] void get_merge_coords
	(char *mc,int *sR,int *sC,int *eR,int *eC)
{
	int i;
	char **comp;
	*sR= 0; *sC=0; *eR= 0; *eC= 0;
	comp= strsplit(":", mc);
	if (!comp) 
		return ;
	if (comp[0]) {
		parse_cell_name(comp[0],sR,sC);
		free(comp[0]);
	} else goto end;

	if (comp[1]) {
		parse_cell_name(comp[1],eR,eC);
		free(comp[1]);
	} else goto end;
	
	for(i=2;comp[i];i++)
		free(comp[i]);
end:
	free(comp);
}

-[private] void parse_cell_name(char *cn,int *row,int *col)
{
	char c;
	int i;
	for(i=0;cn[i]>='A' && cn[i]<='Z';i++)
		;
	if (cn[i]) *row= atoi(cn+i)-1;
	c= cn[i];
	cn[i]= 0;
	*col= parse_col_name(cn);
	cn[i]= c;
}

-[private] int parse_col_name(char *x) {
	int S;
	S= 0;
	while(*x) {
		S= S*26 + x[0]-'A';
		x++;
	}
	return S;
}


-[ headers, begin]
	!stdio !stdlib !string
	parsedoc
-[end]

-[import strsplit function in private scope ]


-[private] void do_escape_sequences(char *X)
{
	int i,j;
	j= 0;
	for(i=0;X[i];i++) {
		if (X[i]!='!' && X[i]!='\\') { X[j++]=X[i]; continue; }
		if (X[i]=='!') { X[j++]= '\n'; continue; }
	}
	X[j]= 0;
}